from datetime import date
from typing import Optional, Union
from urllib.parse import urljoin

from django.conf import settings
from django.dispatch import receiver
from django.urls import reverse
from django.utils.formats import date_format
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from jsonstore import BooleanField
from reversion.models import Revision

from aleksis.apps.chronos.models import (
    Event,
    ExtraLesson,
    LessonSubstitution,
    SupervisionSubstitution,
)
from aleksis.core.models import Announcement, Group, Notification, Person

from .managers import TimetableType
from .models import Lesson, LessonPeriod, Subject
from .util.change_tracker import timetable_data_changed


@Person.property_
def is_teacher(self):
    """Check if the user has lessons as a teacher."""
    return self.lesson_periods_as_teacher.exists()


@Person.property_
def timetable_type(self) -> Optional[TimetableType]:
    """Return which type of timetable this user has."""
    if self.is_teacher:
        return TimetableType.TEACHER
    elif self.primary_group:
        return TimetableType.GROUP
    else:
        return None


@Person.property_
def timetable_object(self) -> Optional[Union[Group, Person]]:
    """Return the object which has the user's timetable."""
    type_ = self.timetable_type

    if type_ == TimetableType.TEACHER:
        return self
    elif type_ == TimetableType.GROUP:
        return self.primary_group
    else:
        return None


@Person.property_
def lessons_as_participant(self):
    """Return a `QuerySet` containing all `Lesson`s this person participates in (as student).

    .. note:: Only available when AlekSIS-App-Chronos is installed.

    :Date: 2019-11-07
    :Authors:
        - Dominik George <dominik.george@teckids.org>
    """
    return Lesson.objects.filter(groups__members=self)


@Person.property_
def lesson_periods_as_participant(self):
    """Return a `QuerySet` containing all `LessonPeriod`s this person participates in (as student).

    .. note:: Only available when AlekSIS-App-Chronos is installed.

    :Date: 2019-11-07
    :Authors:
        - Dominik George <dominik.george@teckids.org>
    """
    return LessonPeriod.objects.filter(lesson__groups__members=self)


@Person.property_
def lesson_periods_as_teacher(self):
    """Return a `QuerySet` containing all `Lesson`s this person gives (as teacher).

    .. note:: Only available when AlekSIS-App-Chronos is installed.

    :Date: 2019-11-07
    :Authors:
        - Dominik George <dominik.george@teckids.org>
    """
    return LessonPeriod.objects.filter(lesson__teachers=self)


@Person.method
def lessons_on_day(self, day: date):
    """Get all lessons of this person (either as participant or teacher) on the given day."""
    qs = LessonPeriod.objects.on_day(day).filter_from_person(self)
    if qs:
        # This is a union queryset, so order by must be after the union.
        return qs.order_by("period__period")
    return None


@Person.method
def _adjacent_lesson(
    self, lesson_period: "LessonPeriod", day: date, offset: int = 1
) -> Union["LessonPeriod", None]:
    """Get next/previous lesson of the person (either as participant or teacher) on the same day."""
    daily_lessons = self.lessons_on_day(day)

    if not daily_lessons:
        return None

    ids = list(daily_lessons.values_list("id", flat=True))

    # Check if the lesson period is one of the person's lesson periods on this day
    # and return None if it's not so
    if lesson_period.pk not in ids:
        return None

    index = ids.index(lesson_period.pk)

    if (offset > 0 and index + offset < len(ids)) or (offset < 0 and index >= -offset):
        return daily_lessons[index + offset]
    else:
        return None


@Person.method
def next_lesson(self, lesson_period: "LessonPeriod", day: date) -> Union["LessonPeriod", None]:
    """Get next lesson of the person (either as participant or teacher) on the same day."""
    return self._adjacent_lesson(lesson_period, day)


@Person.method
def previous_lesson(self, lesson_period: "LessonPeriod", day: date) -> Union["LessonPeriod", None]:
    """Get previous lesson of the person (either as participant or teacher) on the same day."""
    return self._adjacent_lesson(lesson_period, day, offset=-1)


def for_timetables(cls):
    """Return all announcements that should be shown in timetable views."""
    return cls.objects.filter(show_in_timetables=True)


Announcement.class_method(for_timetables)
Announcement.field(
    show_in_timetables=BooleanField(verbose_name=_("Show announcement in timetable views?"))
)

Group.foreign_key("subject", Subject, related_name="groups")

# Dynamically add extra permissions to Group and Person models in core
# Note: requires migrate afterwards
Group.add_permission(
    "view_group_timetable",
    _("Can view group timetable"),
)
Person.add_permission(
    "view_person_timetable",
    _("Can view person timetable"),
)


@receiver(timetable_data_changed)
def send_notifications(sender: Revision, **kwargs):
    """Send notifications to users about the changes."""
    for change in sender.changes.values():
        if change.deleted:
            continue

        recipients = []
        if isinstance(change.instance, LessonSubstitution):
            recipients += change.instance.lesson_period.lesson.teachers.all()
            recipients += change.instance.teachers.all()
            recipients += Person.objects.filter(
                member_of__in=change.instance.lesson_period.lesson.groups.all()
            )
        elif isinstance(change.instance, Event):
            recipients += change.instance.teachers.all()
            recipients += Person.objects.filter(member_of__in=change.instance.groups.all())
        elif isinstance(change.instance, ExtraLesson):
            recipients += change.instance.teachers.all()
            recipients += Person.objects.filter(member_of__in=change.instance.groups.all())
        elif isinstance(change.instance, SupervisionSubstitution):
            recipients.append(change.instance.teacher)
            recipients.append(change.instance.supervision.teacher)
        print(change.instance, recipients)

        description = ""
        if isinstance(change.instance, LessonSubstitution):
            # Date, lesson, subject
            subject = change.instance.lesson_period.lesson.subject
            day = change.instance.date
            period = change.instance.lesson_period.period

            if change.instance.cancelled:
                description += (
                    _(
                        "The {subject} lesson in the {period}. period on {day} has been cancelled."
                    ).format(subject=subject.name, period=period.period, day=date_format(day))
                    + " "
                )
            else:
                description += (
                    _(
                        "The {subject} lesson in the {period}. period "
                        "on {day} has some current changes."
                    ).format(subject=subject.name, period=period.period, day=date_format(day))
                    + " "
                )

                if change.instance.teachers.all():
                    description += (
                        ngettext(
                            "The teacher {old} is substituted by {new}.",
                            "The teachers {old} are substituted by {new}.",
                            change.instance.teachers.count(),
                        ).format(
                            old=change.instance.lesson_period.lesson.teacher_names,
                            new=change.instance.teacher_names,
                        )
                        + " "
                    )

                if change.instance.subject:
                    description += (
                        _("The subject is changed to {subject}.").format(
                            subject=change.instance.subject.name
                        )
                        + " "
                    )

                if change.instance.room:
                    description += (
                        _("The lesson is moved from {old} to {new}.").format(
                            old=change.instance.lesson_period.room.name,
                            new=change.instance.room.name,
                        )
                        + " "
                    )

            if change.instance.comment:
                description += (
                    _("There is an additional comment: {comment}.").format(
                        comment=change.instance.comment
                    )
                    + " "
                )

        elif isinstance(change.instance, Event):
            if change.instance.date_start != change.instance.date_end:
                description += (
                    _(
                        "There is an event that starts on {date_start}, {period_from}. period "
                        "and ends on {date_end}, {period_to}. period:"
                    ).format(
                        date_start=date_format(change.instance.date_start),
                        date_end=date_format(change.instance.date_end),
                        period_from=change.instance.period_from.period,
                        period_to=change.instance.period_to.period,
                    )
                    + "\n"
                )
            else:
                description += (
                    _(
                        "There is an event on {date} from the "
                        "{period_from}. period to the {period_to}. period:"
                    ).format(
                        date=date_format(change.instance.date_start),
                        period_from=change.instance.period_from.period,
                        period_to=change.instance.period_to.period,
                    )
                    + "\n"
                )

            if change.instance.groups.all():
                description += (
                    _("Groups: {groups}").format(groups=change.instance.group_names) + "\n"
                )
            if change.instance.teachers.all():
                description += (
                    _("Teachers: {teachers}").format(teachers=change.instance.teacher_names) + "\n"
                )
            if change.instance.rooms.all():
                description += (
                    _("Rooms: {rooms}").format(
                        rooms=", ".join([room.name for room in change.instance.rooms.all()])
                    )
                    + "\n"
                )
        elif isinstance(change.instance, ExtraLesson):
            description += (
                _("There is an extra lesson on {date} in the {period}. period:").format(
                    date=date_format(change.instance.date),
                    period=change.instance.period.period,
                )
                + "\n"
            )

            if change.instance.groups.all():
                description += (
                    _("Groups: {groups}").format(groups=change.instance.group_names) + "\n"
                )
            if change.instance.room:
                description += (
                    _("Subject: {subject}").format(subject=change.instance.subject.name) + "\n"
                )
            if change.instance.teachers.all():
                description += (
                    _("Teachers: {teachers}").format(teachers=change.instance.teacher_names) + "\n"
                )
            if change.instance.room:
                description += _("Room: {room}").format(room=change.instance.room.name) + "\n"
            if change.instance.comment:
                description += (
                    _("Comment: {comment}.").format(comment=change.instance.comment) + "\n"
                )
        elif isinstance(change.instance, SupervisionSubstitution):
            description += _(
                "The supervision of {old} on {date} between the {period_from}. period "
                "and the {period_to}. period in the area {area} is substituted by {new}."
            ).format(
                old=change.instance.supervision.teacher.full_name,
                date=date_format(change.instance.date),
                period_from=change.instance.supervision.break_item.after_period_number,
                period_to=change.instance.supervision.break_item.before_period_number,
                area=change.instance.supervision.area.name,
                new=change.instance.teacher.full_name,
            )

        day = (
            change.instance.date if hasattr(change.instance, "date") else change.instance.date_start
        )
        url = urljoin(
            settings.BASE_URL, reverse("my_timetable_by_date", args=[day.year, day.month, day.day])
        )

        for recipient in recipients:
            n = Notification(
                recipient=recipient,
                sender=_("Timetable"),
                title=_("There are current changes to your timetable."),
                description=description,
                link=url,
            )
            n.save()
